#include "mas35x9_fs.h"
#include "misc.h"


int mas35x9_dir_readdir(struct file * filp, void * dirent, filldir_t filldir)
{
    int i;
    struct dentry *dentry = filp->f_dentry;

	a("mas35x9_dir_readdir");

    i = filp->f_pos;
    switch (i)
	{
        case 0:
            if (filldir(dirent, ".", 1, i, dentry->d_inode->i_ino, DT_DIR) < 0)
                break;
            i++;
            filp->f_pos++;
			break;

        case 1:
            if (filldir(dirent, "..", 2, i, dentry->d_parent->d_inode->i_ino, DT_DIR) < 0)
                break;
            i++;
            filp->f_pos++;
			break;

        default:
		{
            struct list_head *list;
            int j = i-2;

            spin_lock(&dcache_lock);
            list = dentry->d_subdirs.next;

            for (;;) {
                if (list == &dentry->d_subdirs) {
                    spin_unlock(&dcache_lock);
                    return 0;
                }
                if (!j)
                    break;
                j--;
                list = list->next;
            }

            while(1) {
                struct dentry *de = list_entry(list, struct dentry, d_child);

//				p_dentry(de, "mas35x9_dir_readdir");

                if (!list_empty(&de->d_hash) && de->d_inode) {
                    spin_unlock(&dcache_lock);
                    if (filldir(dirent, de->d_name.name, de->d_name.len, filp->f_pos, de->d_inode->i_ino, DT_UNKNOWN) < 0)
                        break;
                    spin_lock(&dcache_lock);
                }
                filp->f_pos++;
                list = list->next;
                if (list != &dentry->d_subdirs)
                    continue;
                spin_unlock(&dcache_lock);
                break;
            }

        }
    }

    return 0;

}

int		mas35x9_file_open(struct inode* inode, struct file* file)
{
	a("mas35x9_file_open()");

//	p_dentry(inode->d_entry, "mas35x9_file_open()");
	


	return 0;
}

int		mas35x9_file_release(struct inode* indoe, struct file* file)
{
	a("mas35x9_file_release()");
	return 0;
}

/*
	poffset is be in kernel address space. so don't use copy_from_user function.
*/
ssize_t mas35x9_file_read(struct file* file, char* usr_buf, size_t count, loff_t* poffset)
{
	struct inode*	inode = file->f_dentry->d_inode;
	ssize_t	rtn_readed = 0;

	loff_t	offset = *poffset;	// copy only for safety.

	char	ab_area[SMC_AB_AREA_SIZE] = {'\0'};

	u32		i=0;
	int		tmp = 0, tmp2 = 0;
	int		tmp_len = 0;

	u16*	idx_table_of_blocks = inode->u.generic_ip;
	u32		idx_real_page;

	u32		start_idx_virtual_page;
	u32		end_idx_virtual_page;

	if(offset >= inode->i_size)
		return 0;

	a("");
	a("mas35x9_file_read");

	start_idx_virtual_page = get_virtual_start_page(offset);

		tmp = MIN(inode->i_size - offset, count)
	end_idx_virtual_page   = get_virtual_end_page(offset, tmp);

	printk("offset : %d \n", offset);
	printk("  tmp  : %d \n", tmp);
	printk("count  : %d \n", count);

//	if(offset+count >= inode->i_size)
//		return -1;
	
	// skip page which store filename's image to display LCD
	// use two page in start block
	start_idx_virtual_page +=2;
	end_idx_virtual_page   +=2;

	printk("start_idx_virtual_page : %d \n", start_idx_virtual_page);
	printk("  end_idx_virtual_page : %d \n", end_idx_virtual_page);

	for(i  = start_idx_virtual_page;
		i  < end_idx_virtual_page; 
		++i)
	{
		// be careful
		idx_real_page = idx_table_of_blocks[i/page_per_block] *page_per_block + (i%page_per_block);

		tmp = smc_read_page_A_area(idx_real_page, ab_area);
		if(tmp <=0){	return -1;	}
		tmp = smc_read_page_B_area(idx_real_page, ab_area +SMC_A_AREA_SIZE);
		if(tmp <=0){	return -1;	}


		if( i == start_idx_virtual_page){
			tmp2    = offset % SMC_AB_AREA_SIZE;
			tmp_len = MIN(SMC_AB_AREA_SIZE-tmp2, count);

			copy_to_user(usr_buf+rtn_readed, (ab_area +tmp2), tmp_len);
			
			rtn_readed += tmp_len;
		}
		else
		{
			if( i == end_idx_virtual_page -1)
			{
				tmp_len = inode->i_size % SMC_AB_AREA_SIZE;
				printk("end_idx_virtual_page -1 : tmp_len  : %d \n", tmp_len);
				if(0 == tmp_len){
					tmp_len = SMC_AB_AREA_SIZE;
					printk("tmp_len == 0\n");
				}
				printk("end_idx_virtual_page -1 : tmp_len  : %d \n", tmp_len);
			}
			else{
				tmp_len = SMC_AB_AREA_SIZE;
			}

			copy_to_user(usr_buf+rtn_readed, ab_area, tmp_len);

			rtn_readed += tmp_len;
		}

		printk("idx virtual page : %d \n", i);
		printk("idx real page    : %d \n", idx_real_page);

		printk("tmp_len  : %d \n", tmp_len);
		printk("?  : %d \n", i/page_per_block);
		printk("idx_table_of_blocks[?] : %d \n", idx_table_of_blocks[i/page_per_block]);

	}// for

//	return 0;
//	return generic_file_read(file, buf, count, poffset);

	*poffset += rtn_readed;

	printk("rtn_readed %d\n", rtn_readed);

	return rtn_readed;
}



loff_t  mas35x9_file_llseek(struct file* file, loff_t offset, int whence)
{
	off_t	file_length = file->f_dentry->d_inode->i_size;
	loff_t	crnt_offset = file->f_pos;

	a("");
	a("mas35x9_file_llseek");

	printk("offset : %d\n", offset);
	printk("whence : %d\n", whence);
	printk("crnt_offset : %d\n", crnt_offset);


	if( (SEEK_SET != whence) && (SEEK_CUR != whence) && (SEEK_END != whence)){
		a("11");
		return FAIL;
	}

	a("pass");

	if(SEEK_SET == whence)
	{
		a("22");

		if(0+offset >= file_length){
			a("22-1");
			return FAIL;
		}
		if(0+offset < 0){
			a("22-2");
			return FAIL;
		}

		file->f_pos = offset;

		return offset;
	}

	if(SEEK_CUR == whence){
		if(crnt_offset+offset >= file_length){
			return FAIL;
		}
		if(crnt_offset+offset < 0){
			return FAIL;
		}

		file->f_pos = crnt_offset+offset;

		return crnt_offset+offset;
	}

	if(SEEK_END == whence){
		if(file_length+offset >= file_length){
			return FAIL;
		}
		if(file_length+offset < 0){
			return FAIL;
		}

		file->f_pos = file_length+offset;

		return file_length+offset;
	}


	return FAIL;
}



int mas35x9_statfs(struct super_block *sb, struct statfs *buf)
{
	a("mas35x9_statfs");

	buf->f_type = MAS35x9_MAGIC;
	buf->f_bsize = PAGE_CACHE_SIZE;
	buf->f_namelen = 255;

	return 0;
}

/* call this function to umount mas35x9_fs. */
void mas35x9_put_super(struct super_block* sb)
{
	int i=0;

	for(i=0; i<NUM_MAS35x9_FILE; ++i)
	{
		if(NULL != mas35x9_file[i])
			kfree(mas35x9_file[i]);
	}
}


struct dentry * mas35x9_lookup(struct inode *dir, struct dentry *dentry)
{
	a("mas35x9_lookup");
//	p_dentry(dentry, "rafms_lookup");
	d_add(dentry, NULL);

	return NULL;
}


struct inode *mas35x9_get_inode(struct super_block *sb, int mode, int dev)
{
	struct inode * inode = new_inode(sb);

	a("mas35x9_get_inode");

	if (inode)
	{
		inode->i_mode = mode;
		inode->i_uid = UID_ROOT;
		inode->i_gid = GID_ROOT;
/*
		inode->i_blksize = SMC_BLOCK_SIZE;
		inode->i_blocks  = inode->i_blksize/SMC_BLOCK_SIZE + (inode->i_blksize % SMC_BLOCK_SIZE) 1: 0;
*/
		inode->i_rdev = NODEV;
		inode->i_atime= inode->i_mtime = inode->i_ctime = (time_t)NULL;

		switch (mode & S_IFMT)
		{
			default:
				init_special_inode(inode, mode, dev);
				break;

			case S_IFREG:	// 
				inode->i_fop = &mas35x9_file_operations;
				break;

			case S_IFDIR:	// use only one time. to make root directory.
				inode->i_op = &mas35x9_inode_operations;
				inode->i_fop = &mas35x9_dir_operations;
				break;
			}
	}
	return inode;
}


int mas35x9_mknod(struct inode *dir, struct dentry *dentry, int mode, int dev)
{
	struct inode * inode = mas35x9_get_inode(dir->i_sb, mode, dev);
	int error = -ENOSPC;

	a("mas35x9_mknod");

//	p_dentry(dentry, "mas35x9_mknod before d_instantiate");

	if (inode) {
		d_instantiate(dentry, inode);
		dget(dentry);		/* Extra count - pin the dentry in core */
		error = 0;
	}

//	p_dentry(dentry, "mas35x9_mknod after d_instantiate");

	return error;
}


int mas35x9_create(struct inode *dir, struct dentry *dentry, int mode)
{
	a("mas35x9_create");

	return mas35x9_mknod(dir, dentry, mode | S_IFREG, 0);
}



static struct super_block *mas35x9_read_super(struct super_block * sb, void * data, int silent)
{
	struct inode*	inode_root;
	struct dentry*	de_root;

	a("mas35x9_read_super()");

	sb->s_blocksize = PAGE_CACHE_SIZE;
	sb->s_blocksize_bits = PAGE_CACHE_SHIFT;
	sb->s_magic = MAS35x9_MAGIC;
	sb->s_op = &mas35x9_sops;

	inode_root = mas35x9_get_inode(sb, S_IFDIR | 0755, 0);

	if (!inode_root)
		return NULL;

	de_root = d_alloc_root(inode_root);
	if (!de_root) {
		iput(inode_root);
		return NULL;
	}
	sb->s_root = de_root;

	get_mas35x9_proc_info();
	page_per_block	= gSMC_info.page_per_block;

//	printk("page_per_block : 0x%4X \n", page_per_block);

	idx_mas35x9_file = 0;
	mas35x9_scan_root_block(de_root);


	return sb;
}


int  mas35x9_scan_root_block(struct dentry*	de_root)
{
	int		len = 0;
	int		i = 0, j=0, k=0;
	int		idx=0;
	int		page_num_of_root_block = 0;

	s8		filename[18+1+3] = {'\0'};
	struct dentry*	de_child;
	struct inode*	ind_child;

	
	void*			p_tmp = NULL;
	MAS35x9_FILE*	tmp_file[32] = { NULL };


	p_tmp = kmalloc(SMC_AB_AREA_SIZE, GFP_KERNEL);


	for(page_num_of_root_block = 0;
		page_num_of_root_block < 4;
		++page_num_of_root_block)
	{
		memset(p_tmp, '\0', SMC_AB_AREA_SIZE);
		smc_read_page_A_area(gSMC_root_block * page_per_block +page_num_of_root_block,	 p_tmp);
		smc_read_page_B_area(gSMC_root_block * page_per_block +page_num_of_root_block,	 p_tmp + SMC_A_AREA_SIZE);

		for(k=0; k< 32; ++k){
			tmp_file[k] = (MAS35x9_FILE*)(p_tmp + (sizeof(MAS35x9_FILE) * k));
		}

	//	dump_8(tmp_file, 512);
		for(i=0; i<32; ++i)	// origianl
//		for(i=0; idx_mas35x9_file < 1; ++i)	// for debug. only one file.
		{
			if( (0 != tmp_file[i]->name[0]) &&
				(-1!= tmp_file[i]->name[0]) &&
				(90!= tmp_file[i]->name[0]) )
			{
				idx = idx_mas35x9_file;
				++idx_mas35x9_file;

				dump_8(tmp_file[i]->name, 18);

				memset(filename, '\0', 18+1+3);

				mas35x9_file[idx] = kmalloc(sizeof(MAS35x9_FILE), GFP_KERNEL);

				if(NULL == mas35x9_file[idx]){
					a("FAIL : mas35x9_file[idx_mas35x9_file] = kmalloc()");
					goto ERROR;
				}

				memset(mas35x9_file[idx], '\0', sizeof(MAS35x9_FILE));
				memcpy(mas35x9_file[idx], tmp_file[i], sizeof(MAS35x9_FILE));

				// get filename length without extention.
				for(j=0; j<18; ++j){
					if(-1 == mas35x9_file[idx]->name[j])
						break;
				}
				len = j;
	//			c("len", len);

				memcpy(filename, mas35x9_file[idx]->name, len);
				filename[len] = '.';

				memcpy(&filename[len+1], mas35x9_file[idx]->extention, 3);

	
				de_child = make_child_file(de_root, filename, mas35x9_file[idx]);

				if( de_child ){
				// creation of child dentry is successful.
					ind_child = de_child->d_inode;
					ind_child->u.generic_ip = &idx_blocks[idx][0];

					a("===<<<");
//					p_mas35x9_file(mas35x9_file[idx]);
					mas35x9_get_file_blocks(ind_child, mas35x9_file[idx]);
					a("===>>>");
				}
				else
				{
					goto ERROR;
				}
			}

			udelay(100);
		}
	}


ERROR:
	kfree(p_tmp);

	return idx_mas35x9_file;
}



/*
struct dentry* make_child_directory(struct dentry* de_parent, const char*	name)
{
	int len = strlen(name);

	struct dentry*	de_child;

	struct qstr qstr_child = {
		name : name,
		len  : len,
		hash : full_name_hash(name, len)
	};

	printk("make_child_directory\n");

	down(&de_parent->d_inode->i_sem);
	de_child = lookup_hash(&qstr_child, de_parent);
	up(&de_parent->d_inode->i_sem);

	if(de_child <= 0){
		return NULL;
	}

	de_parent->d_inode->i_op->mkdir(de_parent->d_inode, de_child, S_IRUGO);


	return de_child;
}
*/


struct dentry* make_child_file(struct dentry* de_parent, const char*	name, MAS35x9_FILE*	mas35x9_file)
{
	struct dentry*	de_child;
	struct inode*	ind_child;

	int len = strlen(name);
	int	error = 0;
	struct qstr qstr_child = {
		name : name,
		len  : len,
		hash : full_name_hash(name, len)
	};

//	p_qstr(&qstr_child, "make_child_dentry");

	down(&de_parent->d_inode->i_sem);
	de_child = lookup_hash(&qstr_child, de_parent);
	up(&de_parent->d_inode->i_sem);

	if(de_child <= 0){
		return NULL;
	}

	error = de_parent->d_inode->i_op->create(de_parent->d_inode, de_child, S_IRUGO);

	ind_child = de_child->d_inode;

	if(OK == error)
	{
		ind_child->i_size = mas35x9_file->size;
		ind_child->i_blksize = SMC_BLOCK_SIZE;

		/* number of blocks to store this mpe file */
		ind_child->i_blocks  = ind_child->i_size / SMC_BLOCK_SIZE;

		if(ind_child->i_size % SMC_BLOCK_SIZE)
			ind_child->i_blocks++;

		return de_child;
	}

	return NULL;
}


int	mas35x9_get_file_blocks(struct inode* inode, MAS35x9_FILE* mas35x9_file)
{
	int		i = 0;
	u16*	blocks   = inode->u.generic_ip;
	u32		idx_page_to_read = (u32)mas35x9_file->start_block * (u32)page_per_block;
	C_AREA	c_area;


	for(i=0; i < inode->i_blocks +1; ++i)
	{
		if(0 == i){
			blocks[i] = mas35x9_file->start_block;
		}
		
		if(i > 0)
		{
			idx_page_to_read = blocks[i-1] * page_per_block;
			memset(&c_area, '\0', sizeof(C_AREA));

			if(IsOK(smc_read_page_C_area(idx_page_to_read, &c_area)))
			{
//				p_c_area(&c_area);
				if( (0x0003 <= c_area.status) && (c_area.status <= 0xffef) )
				{
					blocks[i] = c_area.status;
				}

				switch(c_area.status)
				{
					case STATUS_BAD_BLOCK:
					case STATUS_ROOT_BLOCK:
					case STATUS_EMPTY_BLOCK:
						printk("CRITICAL ERROR \n");
						goto ERROR;
						break;

					case STATUS_NOT_USED_0:
					case STATUS_NOT_USED_1:
					case STATUS_NOT_USED_2:
					case STATUS_NOT_USED_3:
					case STATUS_NOT_USED_4:
					case STATUS_NOT_USED_5:
					case STATUS_NOT_USED_6:
					case STATUS_NOT_USED_7:
						a("CRITICAL ERROR : STATUS_NOT_USED");
						goto ERROR;
						break;

					case STATUS_LAST_BLOCK:
						blocks[i] = STATUS_LAST_BLOCK;
//						a("STATUS_LAST_BLOCK");
						goto OK_END;
						break;
				}
			}
			else{
				a("smc_read_page_C_area ERROR");
				goto ERROR;
			}	// end smc_read_page_C_area()
		}	// end switch
	}// end for

	OK_END:

	for(i=0; i<inode->i_blocks +1; ++i){
		if(STATUS_LAST_BLOCK == blocks[i]){
			printk("%d : LAST Block \n", blocks[i]);
			break;
		}else{
//			printk("%d : %u\n", i, blocks[i]);
		}
	}

//	printk("i_blocks % \n", inode->i_blocks);

	return TRUE;

	ERROR:
		a("ERROR");
		return FAIL;
}


int  mas35x9_fs_init()
{
	return register_filesystem(&mas35x9_fs_type);
}


void mas35x9_fs_exit()
{
	unregister_filesystem(&mas35x9_fs_type);
}

