#include <linux/module.h>
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/slab.h>		// to use : kmalloc, kfree
#include <linux/dcache.h>	// to use : VFS and ????_hash function
#include <linux/delay.h>	// to use : udelay()

#include <asm/semaphore.h>	// to use : down(), up()
#include <asm/uaccess.h>	// to use : copy_to_user, copy_from_user
#include <asm/errno.h>		// to use error code.

/* some random number */
#define MAS35x9_MAGIC	0x19801999


#include "mas35x9_cmd.h"


/*****************************************************************************
	routine for Linux Kernel VFS
*****************************************************************************/
int		mas35x9_file_open(struct inode *, struct file *);
int		mas35x9_file_release(struct inode *, struct file *);
ssize_t mas35x9_file_read(struct file* file, char * buf, size_t count, loff_t *ppos);
loff_t  mas35x9_file_llseek(struct file *, loff_t, int);

static struct file_operations mas35x9_file_operations = {
	open:		mas35x9_file_open,
	release:	mas35x9_file_release,
	read:		mas35x9_file_read,
	llseek:		mas35x9_file_llseek
};


int		mas35x9_dir_readdir(struct file * filp, void * dirent, filldir_t filldir);

static struct file_operations mas35x9_dir_operations = {
	readdir:	mas35x9_dir_readdir,
};


struct inode*		mas35x9_get_inode(struct super_block *sb, int mode, int dev);


int				mas35x9_create(struct inode *dir, struct dentry *dentry, int mode);
struct dentry*	mas35x9_lookup(struct inode *dir, struct dentry *dentry);
int				mas35x9_mknod(struct inode *dir, struct dentry *dentry, int mode, int dev);

static struct inode_operations mas35x9_inode_operations = {
	create:		mas35x9_create,
	lookup:		mas35x9_lookup,
	mknod:		mas35x9_mknod
};


int	 mas35x9_statfs(struct super_block *sb, struct statfs *buf);
void mas35x9_put_super(struct super_block *);

static struct super_operations mas35x9_sops = {
	statfs:		mas35x9_statfs,
	put_inode:	force_delete,
};


static struct super_block *mas35x9_read_super(struct super_block * sb, void * data, int silent);

struct file_system_type mas35x9_fs_type = {
		name:           "mas35x9_fs",
		read_super:     mas35x9_read_super,
		fs_flags:       FS_LITTER,
//		fs_flags:       FS_NO_DCACHE | FS_NO_PRELIM,
		owner:          THIS_MODULE,
};
/****************************************************************************/



struct dentry* make_child_file(struct dentry* de_parent, 
						       const char*	name, MAS35x9_FILE*	mas35x9_file);

/*
struct dentry* make_child_directory(struct dentry* de_parent, 
									const char*	name, MAS35x9_FILE*	mas35x9_file);
*/


/*****************************************************************************
	extern : mas35x9_usb.h
*****************************************************************************/
extern SMC_INFO	gSMC_info;
extern s16		gSMC_root_block;
extern s8		gMAS35x9_firmware_version;

//int _smc_read_page(u32 page, u16 offset, u16 len, void* data);
extern int smc_read_page_A_area(u32 page, void* pdata);	// length of pdata is longger than 256.
extern int smc_read_page_B_area(u32 page, void* pdata);	// length of pdata is longger than 256.
extern int smc_read_page_C_area(u32 page, void* pdata);	// length of pdata is longger than 16.
/****************************************************************************/



/*****************************************************************************
	extern : mas35x9_usb.h
*****************************************************************************/
extern int get_mas35x9_proc_info();
/****************************************************************************/



/*****************************************************************************
	routine for MAS35x9 file system
*****************************************************************************/
#define NUM_MAS35x9_FILE	64

int		idx_mas35x9_file = 0;
MAS35x9_FILE*	mas35x9_file[NUM_MAS35x9_FILE] = { NULL };
u16		idx_blocks[NUM_MAS35x9_FILE][512] = {{(u16)NULL}};

s8	page_per_block;	// number of pages consist of One SMC Block.

int  mas35x9_fs_init();
void mas35x9_fs_exit();


/*
	return value : number of files in this page.
*/
int mas35x9_scan_root_block(struct dentry*	de_root);
int	mas35x9_get_file_blocks(struct inode* inode, MAS35x9_FILE* mas35x9_file);


/*
u16	get_start_block(loff_t	offset){
	return (offset / SMC_BLOCK_SIZE);
}

u16	get_end_block(loff_t	offset, size_t count){
	u16		rtn = 0;

	rtn = (offset + count) / SMC_BLOCK_SIZE;

	if((offset + count) % SMC_BLOCK_SIZE)
		rtn++;

	return rtn;
}
*/

u32	get_virtual_start_page(loff_t offset)
{
	return (offset / SMC_AB_AREA_SIZE);
}

u32 get_virtual_end_page(loff_t	offset, size_t count)
{
	u16		rtn = 0;

	rtn = (offset + count) / SMC_AB_AREA_SIZE;

	printk("offset + count : %d \n", offset + count);
	printk("(offset + count) / SMC_AB_AREA_SIZE : %d \n", (offset + count) / SMC_AB_AREA_SIZE);
	printk("(offset + count) \% SMC_AB_AREA_SIZE : %d \n", (offset + count) % SMC_AB_AREA_SIZE);

	if( (offset + count) % SMC_AB_AREA_SIZE )
	{
		rtn++;
		printk(" inc rtn\n");
	}

	return rtn;
}




/****************************************************************************/

void p_c_area(C_AREA* c_area)
{
	printk("__c_area \n");
	printk("   status   : 0x%4X \n", c_area->status);
	printk("   attribute: 0x%4X \n", c_area->attribute);
}


void p_mas35x9_file(MAS35x9_FILE* mas35x9_file)
{
	printk("__mas35x9_file \n");
	printk("   attribute   : 0x%4X \n", mas35x9_file->attribute);
	printk("   start_block : 0x%4X \n", mas35x9_file->start_block);
	printk("   size        : 0x%4X \n", mas35x9_file->size);
}


/*	to use for debugging.	*/
void p_dentry(struct dentry* dentry, char *s)
{
	if(s != NULL)
		printk("- dentry [%s] start -\n", s);
	else
		printk("- dentry start -\n");

	printk("   d_count : %d\n", dentry->d_count);
	printk("   d_inode :p %d\n", dentry->d_inode);
	printk("   d_parent:p %d\n", dentry->d_parent);
	printk("\n");
	printk("   d_name.name : %s\n", dentry->d_name.name);
	printk("   d_name.led : %d\n", dentry->d_name.len);
	printk("   d_name.hash : %d\n", dentry->d_name.hash);
	printk("\n");
	printk("   d_name->d_parent.name : %s\n", dentry->d_parent->d_name.name);
	printk("   d_name->d_parent.led : %d\n", dentry->d_parent->d_name.len);
	printk("   d_name->d_parent.hash : %d\n", dentry->d_parent->d_name.hash);
	printk("- dentry end -\n");
}

/*	to use for debugging.	*/
void p_qstr(struct qstr* qstr, char *s)
{
	if(s != NULL)
		printk("- dentry [%s] start -\n", s);
	else
		printk("- dentry start -\n");
	
	printk("qstr->name : %s\n", qstr->name);
	printk("qstr->led : %d\n",  qstr->len);
	printk("qstr->hash : %d\n", qstr->hash);
	printk("- dentry end -\n");
};


/*	to use for debugging.	*/
void dump_8(void* data, int len)
{
	int i = 0;
	int j = 0;
	int limit = 0;

	char*	p = (char*)data;

	
	for(j=0; j<(len/8); ++j)
	{
		limit = (j+1)*8 < len? (j+1)*8 : len;

		for(i= j*8; i<limit ; ++i)
		{
			printk("%4d ", p[i]);
		}

		printk("     ");

		printk("[");
		for(i= j*8; i<limit; ++i)
		{
			switch(p[i])
			{
				case '\0':
					printk(" ");
				break;

				default:
					printk("%c", p[i]);
				break;
			}

		}
		printk("]");

		printk("\n");
	}
}




