#include "mas35x9_proc.h"
#include "misc.h"

int	proc_init()
{
	proc_de_MAS35x9 = create_proc_entry("MAS35x9", S_IFDIR | S_IRUGO, NULL);	

	if(proc_de_MAS35x9 == NULL)
		return FALSE;

	proc_de_info = create_proc_entry("info", S_IFREG | S_IWUGO, proc_de_MAS35x9);

	if(proc_de_info == NULL){
		remove_proc_entry("MAS35x9", NULL);

		return FALSE;
	}

	proc_de_info->proc_fops = &info_fops;

	/*
	proc_de_tmp = create_proc_entry("tmp", S_IFREG | S_IWUGO, proc_de_MAS35x9);
	if(proc_de_tmp == NULL){
		remove_proc_entry("info", NULL);
		remove_proc_entry("MAS35x9", NULL);

		return FALSE;
	}
	proc_de_tmp->write_proc = proc_tmp_write;
	*/

	return TRUE;
}


void proc_exit()
{
	remove_proc_entry("info",	NULL);
	remove_proc_entry("MAS35x9",	NULL);
}


ssize_t info_read(struct file* info_file, char* buf, size_t count, loff_t * offset)
{
	static int bFirst = 0;
	int length_to_copy = -1;

	length_to_copy = get_mas35x9_proc_info();

	if(-1 == length_to_copy)
	{
		printk("-1\n");
		return -1;
	}
	else
	{
		/*
			most of user application, such as more, tail, its buf size is 4096.
			so enough to copy all data at once.
		*/
		if(bFirst ==0){	// first call info_read()
			copy_to_user(buf, proc_info_str, length_to_copy);
			bFirst = 1;
		}
		else{			// second call info_read()
			bFirst = 0;
			return -1;
		}
	}

	return length_to_copy;
}


int get_mas35x9_proc_info()
{
	s8		MAS35x9_firmware_version = 0;
	s32		MAS35x9_buffersize = 0;

	SMC_INFO	SMC_info = {0, 0, 0, 0};
	s16			SMC_root_block = 0;

	int pos = 0;

	/*	VENDOR_CMD_GET_FIRMWARE_VERSION,1,
        VENDOR_CMD_GET_BUFFER_SIZE,     4,
        VENDOR_CMD_GET_SMC_INFO,	    5,
        VENDOR_CMD_GET_ROOT_BLOCK,		2,	*/
	memset(proc_info_str, '\0', PROC_INFO_STR_SIZE);

	mas35x9_vendor_cmd_read(VENDOR_CMD_GET_FIRMWARE_VERSION,	1,	&MAS35x9_firmware_version);
	mas35x9_vendor_cmd_read(VENDOR_CMD_GET_BUFFER_SIZE,	4,	&MAS35x9_buffersize);
	mas35x9_vendor_cmd_read(VENDOR_CMD_GET_SMC_INFO,	5,	&SMC_info);
	mas35x9_vendor_cmd_read(VENDOR_CMD_GET_ROOT_BLOCK,	2,	&SMC_root_block);

	memcpy(&gMAS35x9_firmware_version,
		   &MAS35x9_firmware_version,
		   sizeof(MAS35x9_firmware_version));
	memcpy(&gSMC_info,		 &SMC_info,			sizeof(SMC_info));
	memcpy(&gSMC_root_block, &SMC_root_block,	sizeof(SMC_root_block));

	pos  = sprintf(proc_info_str, "MAS35x9 Firmware Version : %d\n", MAS35x9_firmware_version);
	pos += sprintf(proc_info_str +pos, "MAS35x9 buffer size : %d\n", MAS35x9_buffersize);

	pos += sprintf(proc_info_str +pos, "SMC Manufacturer : (0x%X)", gSMC_info.manufacturer);
	switch(gSMC_info.manufacturer){
		case SMC_SAMSUNG_MANUFACTURER:
			pos += sprintf(proc_info_str +pos, " SAMSUNG ELECTRONICS\n");
			break;

		default:
			pos += sprintf(proc_info_str +pos, " \n");
			break;
	}

	pos += sprintf(proc_info_str +pos, "SMC Size : (0x%X)", gSMC_info.size);
	switch(gSMC_info.size){
		case SMC_SAMSUNG_64M:
			pos += sprintf(proc_info_str +pos, " 64M\n");
			break;
		case SMC_SAMSUNG_128M:
			pos += sprintf(proc_info_str +pos, " 128M\n");
			break;
		default:
			pos += sprintf(proc_info_str +pos, "\n");
			break;
	}

	pos += sprintf(proc_info_str +pos, "SMC Block Count : %d\n", gSMC_info.block_cnt);
	pos += sprintf(proc_info_str +pos, "SMC Block/Page : %d\n", gSMC_info.page_per_block);
	pos += sprintf(proc_info_str +pos, "SMC Root Block Number : %d\n", gSMC_root_block);

	return pos;
}


int		info_open(struct inode* info_inode,
				  struct file* info_file){
	printk("info_open\n");
	return 0;
}

int		info_release(struct inode* info_inode, 
					 struct file* info_file){
	printk("info_release\n");
	return 0;
}

