#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>		// kmalloc, kfree
#include <linux/usb.h>
#include <linux/proc_fs.h>
#include <linux/fs.h>

#include <asm/uaccess.h>	// copy_from_user

#include "mas35x9_cmd.h"

#define	FALSE	0
#define	TRUE	1


EXPORT_NO_SYMBOLS;

MODULE_AUTHOR("Kim ByungKun (skysign@dreamwiz.com)");
MODULE_DESCRIPTION("MAS35x9 MP3 Player Driver for Linux");
MODULE_LICENSE("GPL");


char mas35x9_name[] = "[MAS35x9 Driver for Linux]";

/* Get a minor range for your devices from the usb maintainer */
#define USB_MAS35x9_MINOR_BASE	1
/* we can have up to this number of device plugged in at once */
#define MAX_DEVICES		1

#define	NOERROR(x) (x)>=0
#define	ERROR(x) (x)<0

#define MAS35x9_USB_VENDOR_ID	0x0471
#define MAS35x9_USB_PRODUCT_ID	0x0888


struct {
	struct usb_device *					udev;
	struct usb_interface *				interface;
	struct usb_interface_descriptor *	iface_desc;

	struct usb_endpoint_descriptor *	ep0_desc;
	struct usb_endpoint_descriptor *	ep1_desc;
	struct usb_endpoint_descriptor *	ep2_desc;
	struct usb_endpoint_descriptor *	ep3_desc;

	int	bConnected;
}	gMAS35x9;


int   __init module_mas35x9_init();
void  __exit module_mas35x9_exit();
void* mas35x9_probe(struct usb_device *udev,  unsigned intf,  const struct usb_device_id *id );
void  mas35x9_disconnect(struct usb_device *udev, void *v);


static struct usb_device_id mas35x9_usb_device_id [] = {
	{ USB_DEVICE(MAS35x9_USB_VENDOR_ID, MAS35x9_USB_PRODUCT_ID),
	  bDeviceClass    :		0x08,
	  bDeviceSubClass :		0x00,
	  bDeviceProtocol :		0x00,
	  bInterfaceClass :		0x08,
	  bInterfaceSubClass :	0x01,
	  bInterfaceProtocol :	0x00
	},
	{ }					/* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, mas35x9_usb_device_id);


static struct usb_driver mas35x9_usb_driver = {
	name:		mas35x9_name,
	probe:		mas35x9_probe,
	disconnect:	mas35x9_disconnect,
	minor:		USB_MAS35x9_MINOR_BASE,
	id_table:	mas35x9_usb_device_id
};



// These Two function need to include "linux/init.h"
module_init( module_mas35x9_init );
module_exit( module_mas35x9_exit );


int process_urb_error(int error);


int mas35x9_vendor_cmd_read(u16 index,	u16 size,	void* data);
int mas35x9_vendor_cmd_write(u16 index,	u16 size,	void* data);


SMC_INFO	gSMC_info = {0, 0, 0, 0};
s16			gSMC_root_block = 0;
s8			gMAS35x9_firmware_version = 0;

int _smc_read_page(u32 page, u16 offset, u16 len, void* data);
int smc_read_page_A_area(u32 page, void* pdata){
	return	_smc_read_page(page, 0, 256, pdata);
}

int smc_read_page_B_area(u32 page, void* pdata){
//	return _smc_read_page(page, 255, 256, pdata);
	return _smc_read_page(page, 256, 256, pdata);
}

int smc_read_page_C_area(u32 page, void* pdata){
	return _smc_read_page(page, 512, 16, pdata);
}



extern int	proc_init();
extern void	proc_exit();



/*****************************************************************************
	extern : mas35x9_proc.h
*****************************************************************************/
extern int  get_mas35x9_proc_info();
/****************************************************************************/



/*****************************************************************************
	extern : mas35x9_fs.h
*****************************************************************************/
extern struct file_system_type mas35x9_fs_type;

extern int  mas35x9_fs_init();
extern void mas35x9_fs_exit();
/****************************************************************************/



void print_endpoint_status(struct usb_endpoint_descriptor* ep, unsigned int endpoint_number)
{
	printk("info of End Point : %d\n", endpoint_number);
	printk("-----------------------------------------\n");
	printk("bLength			: %d\n", ep->bLength);
	printk("bDescriptorType : %d\n", ep->bDescriptorType);

	printk("bEndpointAddress: %d\n", ep->bEndpointAddress);
	switch(ep->bEndpointAddress & 0x80){	// direction
		case 0x00:
			printk("   RCV Device->PC");
			break;

		case 0x80:
			printk("   SND PC->Device");
			break;

		default:
			printk("    bEndpointAddress error \n");
			break;
	}
	printk(" EP Num %d\n", ep->bEndpointAddress & 0x07);

	printk("bmAttributes	: %d\n", ep->bmAttributes);
	switch(ep->bmAttributes & 0x03){
		case 0x00:
			printk("    Control\n");
			break;

		case 0x01:
			printk("    Async\n");
			break;

		case 0x02:
			printk("    Bulk\n");
			break;

		case 0x03:
			printk("    Interrupt\n");
			break;

		default:
			printk("    bmAttributes error \n");
			break;
	}

	printk("wMaxPacketSize	: %d\n", ep->wMaxPacketSize);
	printk("bInterval		: %d\n", ep->bInterval);
	printk("bRefresh		: %d\n", ep->bRefresh);
	printk("bSynchAddress	: %d\n", ep->bSynchAddress);
	printk("=========================================\n");
}


